# pylint: disable=too-many-instance-attributes, super-init-not-called
"""进一步提取消息详细信息的函数"""
import copy
import functools
from typing import List

from .exceptions import ContextTypeError
from .model import EventMsg
from .model import FriendMsg
from .model import GroupMsg
from .utils import EventNames
from .utils import MsgTypes

try:
    import ujson as json
except Exception:
    import json


def copy_ctx(f):
    @functools.wraps(f)
    def i(ctx):
        return f(copy.deepcopy(ctx))
    return i

#############################Event begin########################################


class _EventMsg(EventMsg):
    def _carry_properties(self, ctx: EventMsg):
        self.message = ctx.message
        self.CurrentQQ = ctx.CurrentQQ

        self.data = ctx.data

        self.EventName = ctx.EventName
        self.EventData = ctx.EventData
        self.EventMsg = ctx.EventMsg

        self.Content = ctx.Content
        self.FromUin = ctx.FromUin
        self.MsgSeq = ctx.MsgSeq
        self.MsgType = ctx.MsgType
        self.ToUin = ctx.ToUin
        self.RedBaginfo = ctx.RedBaginfo


class _GroupRevokeEventMsg(_EventMsg):
    """群成员撤回消息事件"""

    def __init__(self, ctx: EventMsg):
        event_data = ctx.EventData
        self.AdminUserID: int = event_data.get('AdminUserID')
        self.GroupID: int = event_data.get('GroupID')
        self.MsgRandom: int = event_data.get('MsgRandom')
        self.MsgSeq: int = event_data.get('MsgSeq')
        self.UserID: int = event_data.get('UserID')
        super()._carry_properties(ctx)


class _GroupExitEventMsg(_EventMsg):
    """群成员退出群聊事件"""

    def __init__(self, ctx: EventMsg):
        self.UserID = ctx.EventData.get('UserID')
        super()._carry_properties(ctx)


class _GroupJoinEventMsg(_EventMsg):
    """某人进群事件"""

    def __init__(self, ctx: EventMsg):
        e_data = ctx.EventData
        self.InviteUin: int = e_data.get('InviteUin')
        self.UserID: int = e_data.get('UserID')
        self.UserName: str = e_data.get('UserName')
        super()._carry_properties(ctx)


class _FriendRevokeEventMsg(_EventMsg):
    """好友撤回消息事件"""

    def __init__(self, ctx: EventMsg):
        self.MsgSeq = ctx.EventData.get('MsgSeq')
        self.UserID = ctx.EventData.get('UserID')
        super()._carry_properties(ctx)


class _FriendDeleteEventMsg(_EventMsg):
    """删除好友事件"""

    def __init__(self, ctx: EventMsg):
        self.UserID: int = ctx.EventData.get('UserID')
        super()._carry_properties(ctx)


class _GroupAdminsysnotifyEventMsg(_EventMsg):
    """QQ群系统消息通知(加群申请在这里面"""

    def __init__(self, ctx: EventMsg):
        edata = ctx.EventData
        self.Type: int = edata.get('Type')  # 事件类型
        self.MsgTypeStr: str = edata.get('MsgTypeStr')  # 消息类型
        self.MsgStatusStr: str = edata.get('MsgStatusStr')  # 消息类型状态
        self.Who: int = edata.get('Who')  # 触发消息的对象
        self.WhoName: int = edata.get('WhoName')  # 触发消息的对象昵称
        self.GroupID: int = edata.get('GroupId')  # 来自群
        self.GroupName: str = edata.get('GroupName')  # 群名
        self.ActionUin: int = edata.get('ActionUin')  # 邀请人(处理人)
        self.ActionName: str = edata.get('ActionName')  # 邀请人(处理人)昵称
        self.ActionGroupCard: str = edata.get('ActionGroupCard')  # 邀请人(处理人)群名片
        self.Action: str = edata.get('Action')  # 加群理由 11 agree 14 忽略 12/21 disagree
        self.Content: int = edata.get('Content')
        super()._carry_properties(ctx)


class _GroupShutEventMsg(_EventMsg):
    """群禁言事件"""

    def __init__(self, ctx: EventMsg):
        self.GroupID: int = ctx.EventData.get('GroupID')
        self.ShutTime: int = ctx.EventData.get('ShutTime')
        self.UserID: int = ctx.EventData.get('UserID')
        super()._carry_properties(ctx)


class _GroupAdminEventMsg(_EventMsg):
    """管理员变更事件"""

    def __init__(self, ctx: EventMsg):
        self.Flag: int = ctx.EventData.get('Flag')
        self.GroupID: int = ctx.EventData.get('GroupID')
        self.UserID: int = ctx.EventData.get('UserID')
        super()._carry_properties(ctx)


@copy_ctx
def refine_group_revoke_event_msg(ctx: EventMsg) -> _GroupRevokeEventMsg:
    """群成员撤回消息事件"""
    if not isinstance(ctx, EventMsg):
        raise ContextTypeError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_GROUP_REVOKE:
        return _GroupRevokeEventMsg(ctx)
    return None


@copy_ctx
def refine_group_exit_event_msg(ctx: EventMsg) -> _GroupExitEventMsg:
    """群成员退出群聊事件"""
    if not isinstance(ctx, EventMsg):
        raise ContextTypeError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_GROUP_EXIT:
        return _GroupExitEventMsg(ctx)
    return None


@copy_ctx
def refine_group_join_event_msg(ctx: EventMsg) -> _GroupJoinEventMsg:
    """某人进群事件"""
    if not isinstance(ctx, EventMsg):
        raise ContextTypeError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_GROUP_JOIN:
        return _GroupJoinEventMsg(ctx)
    return None


@copy_ctx
def refine_friend_revoke_event_msg(ctx: EventMsg) -> _FriendRevokeEventMsg:
    """好友撤回消息事件"""
    if not isinstance(ctx, EventMsg):
        raise ContextTypeError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_FRIEND_REVOKE:
        return _FriendRevokeEventMsg(ctx)
    return None


@copy_ctx
def refine_friend_delete_event_msg(ctx: EventMsg) -> _FriendDeleteEventMsg:
    """删除好友事件"""
    if not isinstance(ctx, EventMsg):
        raise ContextTypeError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_FRIEND_DELETE:
        return _FriendDeleteEventMsg(ctx)
    return None


@copy_ctx
def refine_group_adminsysnotify_event_msg(ctx: EventMsg) -> _GroupAdminsysnotifyEventMsg:
    """加群申请"""
    if not isinstance(ctx, EventMsg):
        raise ContextTypeError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_GROUP_ADMINSYSNOTIFY:
        return _GroupAdminsysnotifyEventMsg(ctx)
    return None


@copy_ctx
def refine_group_shut_event_msg(ctx: EventMsg) -> _GroupShutEventMsg:
    """群禁言事件"""
    if not isinstance(ctx, EventMsg):
        raise ContextTypeError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_GROUP_SHUT:
        return _GroupShutEventMsg(ctx)
    return None


@copy_ctx
def refine_group_admin_event_msg(ctx: EventMsg) -> _GroupAdminEventMsg:
    """管理员变更事件"""
    if not isinstance(ctx, EventMsg):
        raise ContextTypeError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_GROUP_ADMIN:
        return _GroupAdminEventMsg(ctx)
    return None
#############################Event end##########################################


#############################Group start########################################
class _GroupMsg(GroupMsg):
    def _carry_properties(self, ctx: GroupMsg):
        self.message = ctx.message
        self.CurrentQQ = ctx.CurrentQQ

        self.data = ctx.data

        self.FromGroupId: int = ctx.FromGroupId
        self.FromGroupName: str = ctx.FromGroupId
        self.FromUserId: int = ctx.FromUserId
        self.FromNickName: str = ctx.FromNickName
        self.Content: str = ctx.Content
        self.MsgType: str = ctx.MsgType
        self.MsgTime: int = ctx.MsgTime
        self.MsgSeq: int = ctx.MsgSeq
        self.MsgRandom: int = ctx.MsgRandom
        self.RedBaginfo: dict = ctx.RedBaginfo


class _VoiceGroupMsg(_GroupMsg):
    """群语音消息"""

    def __init__(self, ctx: GroupMsg):
        voice_data = json.loads(ctx.Content)
        self.VoiceUrl: str = voice_data['Url']
        self.Tips: str = voice_data['Tips']
        super()._carry_properties(ctx)


class _VideoGroupMsg(_GroupMsg):
    """群视频消息"""

    def __init__(self, ctx: GroupMsg):
        video_data = json.loads(ctx.Content)
        self.ForwordBuf: str = video_data['ForwordBuf']
        self.ForwordField: int = video_data['ForwordField']
        self.Tips: str = video_data['Tips']
        self.VideoMd5: str = video_data['VideoMd5']
        self.VideoSize: str = video_data['VideoSize']
        self.VideoUrl: str = video_data['VideoUrl']
        super()._carry_properties(ctx)


class _PicGroupMsg(_GroupMsg):
    """群图片/表情包消息"""

    def __init__(self, ctx: GroupMsg):
        pic_data = json.loads(ctx.Content)
        self.GroupPic: List[dict] = pic_data['GroupPic']
        self.Tips: str = pic_data['Tips']
        super()._carry_properties(ctx)


class _AtGroupMsg(_GroupMsg):
    def __init__(self, ctx: GroupMsg):
        super()._carry_properties(ctx)


class _RedBagGroupMsg(_GroupMsg):
    """群红包消息"""

    def __init__(self, ctx: GroupMsg):
        redbag_info = ctx.RedBaginfo
        self.RedBag_Authkey: str = redbag_info.get('Authkey')
        self.RedBag_Channel: int = redbag_info.get('Channel')
        self.RedBag_Des: str = redbag_info.get('Des')
        self.RedBag_FromType: int = redbag_info.get('FromType')
        self.RedBag_FromUin: int = redbag_info.get('FromUin')
        self.RedBag_Listid: str = redbag_info.get('Listid')
        self.RedBag_RedType: int = redbag_info.get('RedType')
        self.RedBag_StingIndex: str = redbag_info.get('StingIndex')
        self.RedBag_Tittle: str = redbag_info.get('Tittle')
        self.RedBag_Token_17_2: str = redbag_info.get('Token_17_2')
        self.RedBag_Token_17_3: str = redbag_info.get('Token_17_3')
        super()._carry_properties(ctx)


@copy_ctx
def refine_voice_group_msg(ctx: GroupMsg) -> _VoiceGroupMsg:
    """群语音消息"""
    if not isinstance(ctx, GroupMsg):
        raise ContextTypeError('Expected `GroupMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.VoiceMsg:
        return _VoiceGroupMsg(ctx)
    return None


@copy_ctx
def refine_video_group_msg(ctx: GroupMsg) -> _VideoGroupMsg:
    """群视频消息"""
    if not isinstance(ctx, GroupMsg):
        raise ContextTypeError('Expected `GroupMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.VideoMsg:
        return _VideoGroupMsg(ctx)
    return None


@copy_ctx
def refine_pic_group_msg(ctx: GroupMsg) -> _PicGroupMsg:
    """群图片/表情包消息"""
    if not isinstance(ctx, GroupMsg):
        raise ContextTypeError('Expected `GroupMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.PicMsg:
        return _PicGroupMsg(ctx)
    return None


@copy_ctx
def refine_RedBag_group_msg(ctx: GroupMsg) -> _RedBagGroupMsg:
    """群红包消息"""
    if not isinstance(ctx, GroupMsg):
        raise ContextTypeError('Expected `GroupMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.RedBagMsg:
        return _RedBagGroupMsg(ctx)
    return None
#############################Group end##########################################


#############################Friend start#######################################
class _FriendMsg(FriendMsg):
    def _carry_properties(self, ctx: FriendMsg):
        self.message = ctx.message
        self.CurrentQQ = ctx.CurrentQQ

        self.data = ctx.data

        self.FromUin: int = ctx.FromUin
        self.ToUin: int = ctx.ToUin
        self.MsgType: str = ctx.MsgType
        self.MsgSeq: int = ctx.MsgSeq
        self.Content: str = ctx.Content
        self.RedBaginfo: dict = ctx.RedBaginfo


class _VoiceFriendMsg(_FriendMsg):
    """好友语音消息"""

    def __init__(self, ctx: FriendMsg):
        voice_data = json.loads(ctx.Content)
        self.VoiceUrl: str = voice_data['Url']
        self.Tips: str = voice_data['Tips']
        super()._carry_properties(ctx)


class _VideoFriendMsg(_FriendMsg):
    """好友视频消息"""

    def __init__(self, ctx: FriendMsg):
        video_data = json.loads(ctx.Content)
        self.ForwordBuf: str = video_data['ForwordBuf']
        self.ForwordField: int = video_data['ForwordField']
        self.Tips: str = video_data['Tips']
        self.VideoMd5: str = video_data['VideoMd5']
        self.VideoSize: str = video_data['VideoSize']
        self.VideoUrl: str = video_data['VideoUrl']
        super()._carry_properties(ctx)


class _PicFriendMsg(_FriendMsg):
    """好友图片/表情包消息"""

    def __init__(self, ctx: FriendMsg):
        pic_data = json.loads(ctx.Content)
        self.GroupPic: List[dict] = pic_data['FriendPic']
        self.Tips: str = pic_data['Tips']
        super()._carry_properties(ctx)


class _RedBagFriendMsg(_FriendMsg):
    """好友红包消息"""

    def __init__(self, ctx: FriendMsg):
        redbag_info = ctx.RedBaginfo
        self.RedBag_Authkey: str = redbag_info.get('Authkey')
        self.RedBag_Channel: int = redbag_info.get('Channel')
        self.RedBag_Des: str = redbag_info.get('Des')
        self.RedBag_FromType: int = redbag_info.get('FromType')
        self.RedBag_FromUin: int = redbag_info.get('FromUin')
        self.RedBag_Listid: str = redbag_info.get('Listid')
        self.RedBag_RedType: int = redbag_info.get('RedType')
        self.RedBag_StingIndex: str = redbag_info.get('StingIndex')
        self.RedBag_Tittle: str = redbag_info.get('Tittle')
        self.RedBag_Token_17_2: str = redbag_info.get('Token_17_2')
        self.RedBag_Token_17_3: str = redbag_info.get('Token_17_3')
        super()._carry_properties(ctx)


@copy_ctx
def refine_voice_friend_msg(ctx: FriendMsg) -> _VoiceFriendMsg:
    """好友语音消息"""
    if not isinstance(ctx, FriendMsg):
        raise ContextTypeError('Expected `FriendMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.VoiceMsg:
        return _VoiceFriendMsg(ctx)
    return None


@copy_ctx
def refine_video_friend_msg(ctx: FriendMsg) -> _VideoFriendMsg:
    """好友视频消息"""
    if not isinstance(ctx, FriendMsg):
        raise ContextTypeError('Expected `FriendMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.VideoMsg:
        return _VideoFriendMsg(ctx)
    return None


@copy_ctx
def refine_pic_friend_msg(ctx: FriendMsg) -> _PicFriendMsg:
    """好友图片/表情包消息"""
    if not isinstance(ctx, FriendMsg):
        raise ContextTypeError('Expected `FriendMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.PicMsg:
        return _PicFriendMsg(ctx)
    return None


@copy_ctx
def refine_RedBag_friend_msg(ctx: FriendMsg) -> _RedBagFriendMsg:
    """好友红包消息"""
    if not isinstance(ctx, FriendMsg):
        raise ContextTypeError('Expected `FriendMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.RedBagMsg:
        return _RedBagFriendMsg(ctx)
    return None
#############################Friend end#########################################



from typing import List

from .exceptions import InvalidConfigError
from .utils import check_schema

try:
    import ujson as json
except Exception:
    import json


class _config:
    def __init__(self, c: dict) -> None:
        # 与iotbot 对应的配置, 不存在只能为None
        # ip
        host = c.get('host')
        if host:
            self.host = check_schema(str(host))
        else:
            self.host = None
        # port
        try:
            self.port = int(c.get('port'))
        except Exception:
            self.port = None
        # 群黑名单
        self.group_blacklist: List[int] = c.get('group_blacklist')
        # 好友黑名单
        self.friend_blacklist: List[int] = c.get('friend_blacklist')

        # webhook 相关配置
        # 开关
        self.webhook = bool(c.get('webhook'))
        # 推送地址
        webhook_post_url = c.get('webhook_post_url')
        if webhook_post_url:
            self.webhook_post_url = check_schema(str(webhook_post_url))
        else:
            self.webhook_post_url = None
        # 推送等待延时
        try:
            self.webhook_timeout = int(c.get('webhook_timeout'))
        except Exception:
            self.webhook_timeout = 10


_config_dict = {}
try:
    with open('./.iotbot.json', encoding='utf-8') as f:
        _config_dict = json.load(f)
except FileNotFoundError:
    pass
except json.JSONDecodeError as e:
    raise InvalidConfigError('配置文件不规范') from e

config = _config(_config_dict)

# print('=====config=====')
# print('port: ', config.port)
# print('host: ', config.host)
# print('webhook: ', config.webhook)
# print('webhook_post_url: ', config.webhook_post_url)
# print('webhook_timeout: ', config.webhook_timeout)
# print('================')



# pylint: disable=protected-access
"""封装部分最常用发送操作"""
import base64
import sys

from .action import Action
from .exceptions import ContextTypeError
from .model import FriendMsg
from .model import GroupMsg


def file_to_base64(path):
    with open(path, 'rb') as f:
        content = f.read()
    return base64.b64encode(content).decode()


def Text(text: str,
         at=False):
    """发送文字 经支持群消息和好友消息接收函数内调用
    :param text: 文字内容
    :param at:是否艾特发送该消息的用户
    """
    text = str(text)
    # 查找消息上下文 `ctx`变量
    ctx = None
    f = sys._getframe()
    upper = f.f_back
    upper_locals = upper.f_locals
    if 'ctx' in upper_locals and isinstance(upper_locals['ctx'], (FriendMsg, GroupMsg)):
        ctx = upper_locals['ctx']
    else:
        for v in upper_locals.values():
            if isinstance(v, (GroupMsg, FriendMsg)):
                ctx = v
                break

    if isinstance(ctx, GroupMsg):
        return Action(ctx.CurrentQQ).send_group_text_msg(
            ctx.FromGroupId,
            content=text,
            atUser=ctx.FromUserId if at else 0
        )
    elif isinstance(ctx, FriendMsg):
        return Action(ctx.CurrentQQ).send_friend_text_msg(
            ctx.FromUin,
            text
        )
    else:
        raise ContextTypeError('经支持群消息和好友消息接收函数内调用')


def Picture(pic_url='',
            pic_base64='',
            pic_path='',
            content=''):
    """发送图片 经支持群消息和好友消息接收函数内调用
    :param pic_url: 图片链接
    :param pic_base64: 图片base64编码
    :param pic_path: 图片文件路径
    :param content: 包含的文字消息

    pic_url, pic_base64, pic_path必须给定一项
    """
    assert any([pic_url, pic_base64, pic_path]), '必须给定一项'

    ctx = None
    f = sys._getframe()
    upper = f.f_back
    upper_locals = upper.f_locals
    if 'ctx' in upper_locals and isinstance(upper_locals['ctx'], (FriendMsg, GroupMsg)):
        ctx = upper_locals['ctx']
    else:
        for v in upper_locals.values():
            if isinstance(v, (FriendMsg, GroupMsg)):
                ctx = v
                break

    if isinstance(ctx, GroupMsg):
        if pic_url:
            return Action(ctx.CurrentQQ).send_group_pic_msg(
                ctx.FromGroupId,
                picUrl=pic_url,
                content=content
            )
        elif pic_base64:
            return Action(ctx.CurrentQQ).send_group_pic_msg(
                ctx.FromGroupId,
                picBase64Buf=pic_base64,
                content=content
            )
        elif pic_path:
            return Action(ctx.CurrentQQ).send_group_pic_msg(
                ctx.FromGroupId,
                picBase64Buf=file_to_base64(pic_path),
                content=content
            )
        return None
    elif isinstance(ctx, FriendMsg):
        if pic_url:
            return Action(ctx.CurrentQQ).send_friend_pic_msg(
                ctx.FromUin,
                picUrl=pic_url,
                content=content
            )
        elif pic_base64:
            return Action(ctx.CurrentQQ).send_friend_pic_msg(
                ctx.FromUin,
                picBase64Buf=pic_base64,
                content=content
            )
        elif pic_path:
            return Action(ctx.CurrentQQ).send_friend_pic_msg(
                ctx.FromUin,
                picBase64Buf=file_to_base64(pic_path),
                content=content
            )
        return None
    else:
        raise ContextTypeError('经支持群消息和好友消息接收函数内调用')


def Voice(voice_url='',
          voice_base64='',
          voice_path=''):
    """发送语音 经支持群消息和好友消息接收函数内调用
    :param voice_url: 语音链接
    :param voice_base64: 语音base64编码
    :param voice_path: 语音文件路径

    voice_url, voice_base64, voice_path必须给定一项
    """
    assert any([voice_url, voice_base64, voice_path]), '必须给定一项'

    ctx = None
    f = sys._getframe()
    upper = f.f_back
    upper_locals = upper.f_locals
    if 'ctx' in upper_locals and isinstance(upper_locals['ctx'], (FriendMsg, GroupMsg)):
        ctx = upper_locals['ctx']
    else:
        for v in upper_locals.values():
            if isinstance(v, (GroupMsg, FriendMsg)):
                ctx = v
                break

    if isinstance(ctx, GroupMsg):
        if voice_url:
            return Action(ctx.CurrentQQ).send_group_voice_msg(
                ctx.FromGroupId,
                voiceUrl=voice_url
            )
        elif voice_base64:
            return Action(ctx.CurrentQQ).send_group_voice_msg(
                ctx.FromGroupId,
                voiceBase64Buf=voice_base64
            )
        elif voice_path:
            return Action(ctx.CurrentQQ).send_group_voice_msg(
                ctx.FromGroupId,
                voiceBase64Buf=file_to_base64(voice_path)
            )
        return None
    elif isinstance(ctx, FriendMsg):
        if voice_url:
            return Action(ctx.CurrentQQ).send_friend_voice_msg(
                ctx.FromUin,
                voiceUrl=voice_url
            )
        elif voice_base64:
            return Action(ctx.CurrentQQ).send_friend_voice_msg(
                ctx.FromUin,
                voiceBase64Buf=voice_base64
            )
        elif voice_path:
            return Action(ctx.CurrentQQ).send_friend_voice_msg(
                ctx.FromUin,
                voiceBase64Buf=file_to_base64(voice_path)
            )
        return None
    else:
        raise ContextTypeError('经支持群消息和好友消息接收函数内调用')



import importlib
import os
import re
from pathlib import Path
from types import ModuleType
from typing import Dict
from typing import List

from prettytable import PrettyTable

from .typing import EventMsgReceiver
from .typing import FriendMsgReceiver
from .typing import GroupMsgReceiver

try:
    import ujson as json
except ImportError:
    import json


class Plugin:
    def __init__(self, module: ModuleType):
        self.module = module
        self.name = Path(module.__file__).stem[4:]
        self.receive_group_msg: GroupMsgReceiver = module.__dict__.get('receive_group_msg')
        self.receive_friend_msg: FriendMsgReceiver = module.__dict__.get('receive_friend_msg')
        self.receive_events: EventMsgReceiver = module.__dict__.get('receive_events')


class PluginManager:
    def __init__(self, plugin_dir: str = 'plugins'):
        self.plugin_dir = plugin_dir
        self._plugins: Dict[str, Plugin] = dict()
        self._removed_plugins: Dict[str, Plugin] = dict()

        # 本地缓存的停用的插件名称列表
        self._load_removed_plugin_names()

    def _load_removed_plugin_names(self):
        if os.path.exists('.REMOVED_PLUGINS'):
            with open('.REMOVED_PLUGINS', encoding='utf8') as f:
                self._removed_plugin_names = json.load(f)['plugins']
        else:
            with open('.REMOVED_PLUGINS', 'w', encoding='utf8') as f:
                json.dump({'tips': '用于存储已停用插件信息,请不要修改这个文件', 'plugins': []}, f, ensure_ascii=False)
            self._removed_plugin_names = []

    def _update_removed_plugin_names(self):
        data = {
            'tips': '用于存储已停用插件信息,请不要修改这个文件',
            'plugins': list(set(self._removed_plugin_names))  # 去重，虽然显得多余
        }
        with open('.REMOVED_PLUGINS', 'w', encoding='utf8') as f:
            json.dump(data, f, ensure_ascii=False)

    def load_plugins(self, plugin_dir: str = None) -> None:
        if plugin_dir is None:
            plugin_dir = self.plugin_dir
        plugin_files = (i for i in os.listdir(plugin_dir) if re.search(r'^bot_\w+\.py$', i))
        for plugin_file in plugin_files:
            module = importlib.import_module(
                '{}.{}'.format(plugin_dir.replace('/', '.'), plugin_file.split('.')[0]))
            plugin = Plugin(module)
            if plugin.name in self._removed_plugin_names:
                self._removed_plugins[plugin.name] = plugin
            else:
                self._plugins[plugin.name] = plugin

    def refresh(self, plugin_dir: str = None) -> None:
        '''reload all plugins'''
        self._plugins.clear()
        self.load_plugins(plugin_dir)
        self.reload_plugins(plugin_dir)

    def reload_plugins(self, plugin_dir: str = None) -> None:
        '''reload old, load new.'''
        # reload old
        old_plugins = self._plugins.copy()
        for old_plugin in old_plugins.values():
            old_plugins[old_plugin.name].module = importlib.reload(old_plugin.module)
        # load new
        self.load_plugins(plugin_dir)
        # tidy
        self._plugins.update(old_plugins)

    def reload_plugin(self, plugin_name: str) -> None:
        '''reload one plugin according to plugin name
        whether the plugin exists or not, it will always keep quiet.
        '''
        if plugin_name in self._plugins:
            self._plugins[plugin_name].module = importlib.reload(self._plugins[plugin_name].module)

    def remove_plugin(self, plugin_name: str) -> None:
        '''remove not delete.'''
        try:
            if plugin_name in self._plugins:
                self._removed_plugins[plugin_name] = self._plugins.pop(plugin_name)
                # 缓存到本地
                self._removed_plugin_names.append(plugin_name)
                self._update_removed_plugin_names()
        except KeyError:  # 可能由self._removed_plugins[plugin_name]引发
            pass

    def recover_plugin(self, plugin_name: str) -> None:
        '''recover plugin if it's in the removed plugins list.'''
        try:
            if plugin_name in self._removed_plugins:
                self._plugins[plugin_name] = self._removed_plugins.pop(plugin_name)
                if plugin_name in self._removed_plugin_names:
                    self._removed_plugin_names.remove(plugin_name)
                    self._update_removed_plugin_names()
        except KeyError:
            pass

    @property
    def plugins(self) -> List[str]:
        '''return a list of plugin name'''
        return list(self._plugins)

    @property
    def removed_plugins(self) -> List[str]:
        '''return a list of removed plugin name'''
        return list(self._removed_plugins)

    @property
    def friend_msg_receivers(self) -> List[FriendMsgReceiver]:
        '''funcs to handle (friend msg)context'''
        return [plugin.receive_friend_msg for plugin in self._plugins.values()
                if plugin.receive_friend_msg]

    @property
    def group_msg_receivers(self) -> List[GroupMsgReceiver]:
        '''funcs to handle (group msg)context'''
        return [plugin.receive_group_msg for plugin in self._plugins.values()
                if plugin.receive_group_msg]

    @property
    def event_receivers(self) -> List[EventMsgReceiver]:
        '''funcs to handle (event msg)context'''
        return [plugin.receive_events for plugin in self._plugins.values()
                if plugin.receive_events]

    @property
    def info_table(self) -> str:
        table = PrettyTable(['Receiver', 'Count', 'Info'])
        table.add_row([
            'Friend Msg Receiver',
            len(self.friend_msg_receivers),
            ' '.join([f'<{p.name}>' for p in self._plugins.values() if p.receive_friend_msg])
        ])
        table.add_row([
            'Group  Msg Receiver',
            len(self.group_msg_receivers),
            ' '.join([f'<{p.name}>' for p in self._plugins.values() if p.receive_group_msg])
        ])
        table.add_row([
            'Event      Receiver',
            len(self.event_receivers),
            ' '.join([f'<{p.name}>' for p in self._plugins.values() if p.receive_events])
        ])
        table_removed = PrettyTable(['Removed Plugins'])
        table_removed.add_row([' / '.join(self.removed_plugins)])
        return str(table) + '\n' + str(table_removed)



import copy
import sys
from collections.abc import Sequence
from concurrent.futures import ThreadPoolExecutor
from typing import Callable
from typing import List
from typing import Union

import socketio

from .config import config
from .logger import logger
from .model import EventMsg
from .model import FriendMsg
from .model import GroupMsg
from .model import model_map
from .plugin import PluginManager
from .typing import EventMsgReceiver
from .typing import FriendMsgReceiver
from .typing import GroupMsgReceiver


def _deco_creater(bind_type):
    def deco(self, func):
        if bind_type == 'OnGroupMsgs':
            self.add_group_msg_receiver(func)
        elif bind_type == 'OnFriendMsgs':
            self.add_friend_msg_receiver(func)
        else:
            self.add_event_receiver(func)
    return deco


class IOTBOT:  # pylint: disable = too-many-instance-attributes
    """
    :param qq: 机器人QQ号(多Q就传qq号列表)
    :param use_plugins: 是否开启插件功能
    :param plugin_dir: 插件存放目录
    :param group_blacklist: 群黑名单, 此名单中的群聊消息不会被处理,默认为空，即全部处理
    :param friend_whitelist: 好友白名单，只有此名单中的好友消息才会被处理，默认为空，即全部处理
    :param log: 是否开启日志
    :param log_file: 是否文件日志
    :param port: 运行端口
    :param beat_delay: 心跳延时时间（s）
    :param host: ip，需要包含schema
    """

    def __init__(self,
                 qq: Union[int, List[int]],
                 use_plugins: bool = False,
                 plugin_dir: str = 'plugins',
                 group_blacklist: List[int] = None,
                 friend_blacklist: List[int] = None,
                 log: bool = True,
                 log_file: bool = True,
                 port: int = 8888,
                 host: str = 'http://127.0.0.1'):
        if isinstance(qq, Sequence):
            self.qq = list(qq)
        else:
            self.qq = [qq]
        self.use_plugins = use_plugins
        self.plugin_dir = plugin_dir
        self.host = config.host or host
        self.port = config.port or port
        self.group_blacklist = set(config.group_blacklist or group_blacklist or [])
        self.friend_blacklist = set(config.friend_blacklist or friend_blacklist or [])

        if log:
            if log_file:
                logger.add(
                    './logs/{time}.log',
                    format='{time:YYYY-MM-DD HH:mm} {level}\t{message}',
                    rotation='1 day',
                    encoding='utf-8'
                )
        else:
            logger.disable(__name__)

        # 手动添加的消息接收函数
        self.__friend_msg_receivers_from_hand = []
        self.__group_msg_receivers_from_hand = []
        self.__event_receivers_from_hand = []

        # webhook 里的消息接收函数，是特例
        if config.webhook:
            from . import webhook  # pylint:disable=import-outside-toplevel
            # 直接加载进 `hand`
            self.__friend_msg_receivers_from_hand.append(webhook.receive_friend_msg)
            self.__group_msg_receivers_from_hand.append(webhook.receive_group_msg)
            self.__event_receivers_from_hand.append(webhook.receive_events)

        # 消息上下文对象中间件
        self.__friend_context_middleware: Callable[[FriendMsg], FriendMsg] = None
        self.__group_context_middleware: Callable[[GroupMsg], GroupMsg] = None
        self.__event_context_middleware: Callable[[EventMsg], EventMsg] = None

        # 插件管理
        self.plugMgr = PluginManager(self.plugin_dir)
        if use_plugins:
            self.plugMgr.load_plugins()
            print(self.plugin_status)

        # 依次各种初始化
        self.__initialize_socketio()
        self.__refresh_executor()
        self.__initialize_handlers()

    ########################################################################
    # shortcuts to call plugin manager methods
    ########################################################################
    # 只推荐使用这几个方法，其他的更细致的方法需要通过 plugMgr 对象访问
    def load_plugins(self):
        '''加载新插件'''
        self.plugMgr.load_plugins()

    def reload_plugins(self):
        '''重载旧插件，加载新插件'''
        self.plugMgr.reload_plugins()

    def reload_plugin(self, plugin_name: str):
        '''重载指定插件'''
        self.plugMgr.reload_plugin(plugin_name)

    def refresh_plugins(self):
        '''刷新插件目录所有插件'''
        self.plugMgr.refresh()

    def remove_plugin(self, plugin_name: str):
        '''停用指定插件'''
        self.plugMgr.remove_plugin(plugin_name)

    def recover_plugin(self, plugin_name: str):
        '''启用指定插件'''
        self.plugMgr.recover_plugin(plugin_name)

    @property
    def plugin_status(self):
        '''插件启用状态'''
        return self.plugMgr.info_table

    @property
    def plugins(self):
        '''插件名列表'''
        return self.plugMgr.plugins

    @property
    def removed_plugins(self):
        '''已停用的插件名列表'''
        return self.plugMgr.removed_plugins
    ########################################################################

    def run(self):
        logger.info('Connecting to the server...')

        try:
            self.socketio.connect(f'{self.host}:{self.port}', transports=['websocket'])
        except Exception:
            logger.exception('启动失败')
            sys.exit(1)
        else:
            try:
                self.socketio.wait()
            except KeyboardInterrupt:
                self.__executor.shutdown(wait=False)
                print('\nbye~')
                sys.exit(0)

    def connect(self):
        logger.success('Connected to server successfully!')
        for qq in self.qq:
            self.socketio.emit(
                'GetWebConn', str(qq),
                callback=lambda x: logger.info(f'GetWebConn -> {qq} => {x}')  # pylint: disable=cell-var-from-loop
            )

    @property
    def receivers(self):
        '''消息处理函数数量'''
        return {
            'friend': len((*self.plugMgr.friend_msg_receivers,
                           *self.__friend_msg_receivers_from_hand)),
            'group': len((*self.plugMgr.group_msg_receivers,
                          *self.__group_msg_receivers_from_hand)),
            'event': len((*self.plugMgr.event_receivers,
                          *self.__event_receivers_from_hand))
        }

    @receivers.setter
    def receivers(self, _):
        logger.warning('The attribute receivers is read-only!')

    def __initialize_socketio(self):
        self.socketio = socketio.Client()
        self.socketio.event()(self.connect)

    def __refresh_executor(self):
        # 根据消息接收函数数量初始化线程池
        self.__executor = ThreadPoolExecutor(max_workers=min(50, len([  # 减小数量，控制消息频率
            *self.plugMgr.friend_msg_receivers,
            *self.__friend_msg_receivers_from_hand,
            *self.plugMgr.group_msg_receivers,
            *self.__group_msg_receivers_from_hand,
            *self.plugMgr.event_receivers,
            *self.__event_receivers_from_hand,
            *range(3)
        ]) * 2))

    # 手动添加
    def add_group_msg_receiver(self, func: GroupMsgReceiver):
        '''添加群消息接收函数'''
        self.__group_msg_receivers_from_hand.append(func)

    def add_friend_msg_receiver(self, func: FriendMsgReceiver):
        '''添加好友消息接收函数'''
        self.__friend_msg_receivers_from_hand.append(func)

    def add_event_receiver(self, func: EventMsgReceiver):
        '''添加事件消息接收函数'''
        self.__event_receivers_from_hand.append(func)

    ########################################################################
    # context distributor
    ########################################################################
    def __friend_context_distributor(self, context: FriendMsg):
        for f_receiver in [*self.__friend_msg_receivers_from_hand,
                           *self.plugMgr.friend_msg_receivers]:
            (self.__executor
             .submit(f_receiver, copy.deepcopy(context))
             .add_done_callback(self.__thread_pool_callback))

    def __group_context_distributor(self, context: GroupMsg):
        for g_receiver in [*self.__group_msg_receivers_from_hand,
                           *self.plugMgr.group_msg_receivers]:
            (self.__executor
             .submit(g_receiver, copy.deepcopy(context))
             .add_done_callback(self.__thread_pool_callback))

    def __event_context_distributor(self, context: EventMsg):
        for e_receiver in [*self.__event_receivers_from_hand,
                           *self.plugMgr.event_receivers]:
            (self.__executor
             .submit(e_receiver, copy.deepcopy(context))
             .add_done_callback(self.__thread_pool_callback))

    ########################################################################
    # register context middleware
    ########################################################################
    def register_friend_context_middleware(self, middleware: Callable[[FriendMsg], FriendMsg]):
        """注册好友消息中间件"""
        if self.__friend_context_middleware is not None:
            raise Exception('Cannot register more than one middleware(friend)')
        self.__friend_context_middleware = middleware

    def register_group_context_middleware(self, middleware: Callable[[GroupMsg], GroupMsg]):
        """注册群消息中间件"""
        if self.__group_context_middleware is not None:
            raise Exception('Cannot register more than one middleware(group)')
        self.__group_context_middleware = middleware

    def register_event_context_middleware(self, middleware: Callable[[EventMsg], EventMsg]):
        """注册事件消息中间件"""
        if self.__event_context_middleware is not None:
            raise Exception('Cannot register more than one middleware(event)')
        self.__event_context_middleware = middleware

    ########################################################################
    # message handler
    ########################################################################
    @logger.catch
    def __thread_pool_callback(self, worker):
        worker_exception = worker.exception()
        if worker_exception:
            raise worker_exception

    def __friend_msg_handler(self, msg):
        context: FriendMsg = model_map['OnFriendMsgs'](msg)
        logger.info(f'{context.__class__.__name__} ->  {context.data}')
        # 黑名单
        if context.FromUin in self.friend_blacklist:
            return
        # 中间件
        if self.__friend_context_middleware is not None:
            new_context = self.__friend_context_middleware(context)
            if isinstance(new_context, type(context)):
                context = new_context
        self.__executor.submit(self.__friend_context_distributor, context)

    def __group_msg_handler(self, msg):
        context: GroupMsg = model_map['OnGroupMsgs'](msg)
        logger.info(f'{context.__class__.__name__} ->  {context.data}')
        # 黑名单
        if context.FromGroupId in self.group_blacklist:
            return
        # 中间件
        if self.__group_context_middleware is not None:
            new_context = self.__group_context_middleware(context)
            if isinstance(new_context, type(context)):
                context = new_context
        self.__executor.submit(self.__group_context_distributor, context)

    def __event_msg_handler(self, msg):
        context: EventMsg = model_map['OnEvents'](msg)
        logger.info(f'{context.__class__.__name__} ->  {context.data}')
        # 中间件
        if self.__event_context_middleware is not None:
            new_context = self.__event_context_middleware(context)
            if isinstance(new_context, type(context)):
                context = new_context
        self.__executor.submit(self.__event_context_distributor, context)

    def __initialize_handlers(self):
        self.socketio.on('OnGroupMsgs')(self.__group_msg_handler)
        self.socketio.on('OnFriendMsgs')(self.__friend_msg_handler)
        self.socketio.on('OnEvents')(self.__event_msg_handler)
    ###########################################################################
    # decorators
    on_group_msg = _deco_creater('OnGroupMsgs')
    on_friend_msg = _deco_creater('OnFriendMsgs')
    on_event = _deco_creater('OnEvents')

    def __repr__(self):
        return 'IOTBOT <{}> <host-{}> <port-{}>'.format(
            " ".join([str(i) for i in self.qq]),
            self.host,
            self.port
        )



"""Author: wongxy

github.com/xiyaowong/python-iotbot

>>> from iotbot import IOTBOT, GroupMsg, FriendMsg, EventMsg

>>> bot = IOTBOT(your_bot_qq)

>>> @bot.on_group_msg
>>> def group(ctx: GroupMsg):
>>>    pass

>>> @bot.on_friend_msg
>>> def friend(ctx: FriendMsg):
>>>    pass

>>> @bot.on_event
>>> def event(message: EventMsg):
>>>     pass
"""

from .action import Action
from .client import IOTBOT
from .model import EventMsg
from .model import FriendMsg
from .model import GroupMsg
from .model import model_map



from typing import Any
from typing import Callable

from .model import EventMsg
from .model import FriendMsg
from .model import GroupMsg

GroupMsgReceiver = Callable[[GroupMsg], Any]
FriendMsgReceiver = Callable[[FriendMsg], Any]
EventMsgReceiver = Callable[[EventMsg], Any]
